﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;

using ThinkGeo.MapSuite.Core;
using ThinkGeo.MapSuite.MvcEdition;

namespace MapResizing
{
    /// <summary>
    /// Datatype representing a geographic coordindate.
    /// </summary>
    public struct GeoCoordinate
    {
        /// <summary>
        /// X coordinate. (Longitude)
        /// </summary>
        public double X { get; set; }

        /// <summary>
        /// Y coordinate. (Latitude)
        /// </summary>
        public double Y { get; set; }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="x">Longitude.</param>
        /// <param name="y">Latitude.</param>
        public GeoCoordinate(double x, double y)
            : this()
        {
            X = x;
            Y = y;
        }
    }

    public enum Projections
    {
        NAD83 = 4269,
        NAD27 = 4267,
        OSM = 3857,
        WGS84 = 4326
    }

    public static class MappingUtility
    {
        /// <summary>
        /// Projection from WGS84 to OpenStreetMaps spherical mercator.
        /// </summary>
        public static Proj4Projection WGS84toSphMercator { get; private set; }

        /// <summary>
        /// Projection from NAD83 to OpenStreetMaps spherical mercator.
        /// </summary>
        public static Proj4Projection NAD83toSphMercator { get; private set; }

        /// <summary>
        /// Projection from NAD27 to OpentStreetMaps spherical mercator.
        /// </summary>
        public static Proj4Projection NAD27toSphMercator { get; private set; }

        /// <summary>
        /// Projection from OpentStreetMaps spherical mercator to WGS84.
        /// </summary>
        public static Proj4Projection SphMercatorToWGS84 { get; private set; }

        /// <summary>
        /// The default map extent used for ACCESS Web 2.0.
        /// </summary>
        public static RectangleShape DefaultExtent { get; private set; }

        /// <summary>
        /// Default map Width.
        /// </summary>
        public static double DefaultMapWidth { get; private set; }

        /// <summary>
        /// Default map Height.
        /// </summary>
        public static double DefaultMapHeight { get; private set; }

        /// <summary>
        /// Default font used for creating marker labels.
        /// </summary>
        private static GeoFont DefaultFont { get; set; }

        //Used to convert longitude and latitude from integer storage form.
        private const double DataBaseStorageConversionFactor = 1000000.0;

        private const double FeetPerMeter = 3.28084;

        /// <summary>
        /// Static Constructor.
        /// </summary>
        static MappingUtility()
        {
            //Projection from NAD83 to OpenStreetMaps spherical mercator
            NAD83toSphMercator = new Proj4Projection((int)Projections.NAD83, (int)Projections.OSM);

            //Projection from NAD27 to OpentStreetMaps spherical mercator
            NAD27toSphMercator = new Proj4Projection((int)Projections.NAD27, (int)Projections.OSM);

            //Projection from WGS84 to OpenStreetMaps spherical mercator
            WGS84toSphMercator = new Proj4Projection((int)Projections.WGS84, (int)Projections.OSM);

            //Projection from OpentStreetMaps spherical mercator to WGS84
            SphMercatorToWGS84 = new Proj4Projection((int)Projections.OSM, (int)Projections.WGS84);

            DefaultExtent = new RectangleShape(-80.4014, 40.7301, -79.5431, 40.1972);

            DefaultMapHeight = 700;

            DefaultMapWidth = 1550;

            DefaultFont = new GeoFont("Arial", 10, DrawingFontStyles.Bold);
        }


        #region Conversions

        /// <summary>
        /// Converts a map coordinate stored as an integer in the database back to floating point representation by dividing by 1000000.0.
        /// </summary>
        /// <param name="coord">The coordinate from the database to convert.</param>
        /// <returns>The coordinate converted back to floating point representation(double). </returns>
        public static double ConvertDBCoordToMap(int coord)
        {
            return (double)coord / DataBaseStorageConversionFactor;
        }

        /// <summary>
        /// Converts a map coordinate in floating point (double) representation to an integer for storage in the database by multiplying by 1000000.0.
        /// </summary>
        /// <param name="coord">The coordinate to convert.</param>
        /// <returns>The coordinate converted to an integer for storage. </returns>
        public static int ConvertMapCoordToDB(double coord)
        {
            return (int)(coord * DataBaseStorageConversionFactor);
        }

        /// <summary>
        /// Converts meters to feet.
        /// </summary>
        /// <param name="meters">The length in meters.</param>
        /// <returns>The length converted to feet.</returns>
        public static double MetersToFeet(double meters)
        {
            return FeetPerMeter * meters;
        }

        /// <summary>
        /// Converts feet to meters.
        /// </summary>
        /// <param name="feet">The lenght in feet.</param>
        /// <returns>The legnth converted to meters.</returns>
        public static double FeetToMeters(double feet)
        {
            return feet / FeetPerMeter;
        }

        /// <summary>
        /// Creates a GeoCoordinate from a pair of coordinates in database storage form(integer).
        /// </summary>
        /// <param name="lat">Latitude.</param>
        /// <param name="lon">Longitude.</param>
        /// <returns>GeoCoordinate created using the values specified.</returns>
        public static GeoCoordinate GeoCoordinateFromDBCoords(int lat, int lon)
        {
            return new GeoCoordinate(ConvertDBCoordToMap(lon), ConvertDBCoordToMap(lat));
        }

        /// <summary>
        /// Creates a PointShape from a pair of coordinates in database storage form(integer).
        /// </summary>
        /// <param name="lat">Latitude.</param>
        /// <param name="lon">Longitude.</param>
        /// <returns>PointShape created using the values specified.</returns>
        public static PointShape PointShapeFromDBCoords(int lat, int lon)
        {
            return new PointShape(ConvertDBCoordToMap(lon), ConvertDBCoordToMap(lat));
        }

        /// <summary>
        /// Creates a Vertex from a pair of coordinates in database storage form(integer).
        /// </summary>
        /// <param name="lat">Latitude.</param>
        /// <param name="lon">Longitude.</param>
        /// <returns>Vertex created using the values specified.</returns>
        public static Vertex VertexFromDBCoords(int lat, int lon)
        {
            return new Vertex(ConvertDBCoordToMap(lon), ConvertDBCoordToMap(lat));
        }

        /// <summary>
        /// Creates a PointShape from a GeoCoordinate.
        /// </summary>
        /// <param name="coords">The GeoCoordinate used for creation of the the PointShpape.</param>
        /// <returns>PointShape created from the GeoCoordinate object.</returns>
        public static PointShape PointShapeFromGeoCoordinate(GeoCoordinate coords)
        {
            return new PointShape(coords.X,coords.Y);
        }

        /// <summary>
        /// Creates a Vertex from a GeoCoordinate.
        /// </summary>
        /// <param name="coords">The GeoCoordinate used for creation of the the Vertex.</param>
        /// <returns>Vertex created from the GeoCoordinate object.</returns>
        public static Vertex VertexFromDBCoords(GeoCoordinate coords)
        {
            return new Vertex(coords.X,coords.Y);
        }

        #endregion

        #region Projections

        /// <summary>
        /// Projects a coordinate pair to the specified Proj4Projection's ExternalProjection.
        /// </summary>
        /// <param name="x">X Coordinate(Longitude).</param>
        /// <param name="y">Y Coordinate(Latitude).</param>
        /// <param name="proj">Proj4Projection to use.</param>
        public static void ProjectCoordinates(ref double x, ref double y, Proj4Projection proj)
        {
            Vertex v = new Vertex();

            try
            {
                proj.Open();

                v = proj.ConvertToExternalProjection(x, y);
            }
            finally
            {
                proj.Close();
            }

            x = v.X;
            y = v.Y;
        }

        /// <summary>
        /// Converts a coordinate pair to a Vertex projected to the specified Proj4Projection's ExternalProjection.
        /// </summary>
        /// <param name="x">X Coordinate(Longitude).</param>
        /// <param name="y">Y Coordinate(Latitude).</param>
        /// <param name="proj">Proj4Projection to use.</param>
        /// <returns>A Vertex consisting of the projected coordinates.</returns>
        public static Vertex ProjectCoordinates(double x, double y, Proj4Projection proj)
        {
            Vertex v = new Vertex();

            try
            {
                proj.Open();

                v = proj.ConvertToExternalProjection(x, y);
            }
            finally
            {
                proj.Close();
            }

            return v;
        }

        /// <summary>
        /// Projects a PointShape to the specified Proj4Projection's ExternalProjection.
        /// </summary>
        /// <param name="p">PointShape to project.</param>
        /// <param name="proj">Proj4Projection to use.</param>
        /// <returns>>A PointShape projected to the specified Proj4Projection's ExternalProjection.</returns>
        public static PointShape ProjectCoordinates(PointShape p, Proj4Projection proj)
        {
            PointShape ps = new PointShape();

            try
            {
                proj.Open();

                ps = (PointShape)proj.ConvertToExternalProjection(p);
            }
            finally
            {
                proj.Close();
            }

            return ps;
        }

        /// <summary>
        /// Projects a PointShape to the specified Proj4Projection's ExternalProjection.
        /// </summary>
        /// <param name="p">PointShape to project.</param>
        /// <param name="proj">Proj4Projection to use.</param>
        public static void ProjectCoordinates(ref PointShape p, Proj4Projection proj)
        {
            PointShape ps = new PointShape();

            try
            {
                proj.Open();

                ps = (PointShape)proj.ConvertToExternalProjection(p);
            }
            finally
            {
                proj.Close();
            }
        }

        /// <summary>
        /// Projects a Vertex to the specified Proj4Projection's ExternalProjection.
        /// </summary>
        /// <param name="v">Vertex to project.</param>
        /// <param name="proj">Proj4Projection to use.</param>
        /// <returns>A Vertex projected to the specified Proj4Projection's ExternalProjection.</returns>
        public static Vertex ProjectCoordinates(Vertex v, Proj4Projection proj)
        {
            Vertex newVert = v;

            try
            {
                proj.Open();

                newVert = proj.ConvertToExternalProjection(v.X, v.Y);
            }
            finally
            {
                proj.Close();
            }

            return newVert;
        }

        /// <summary>
        /// Projects a Vertex to the specified Proj4Projection's ExternalProjection.
        /// </summary>
        /// <param name="v">Vertex to project.</param>
        /// <param name="proj">Proj4Projection to use.</param>
        public static void ProjectCoordinates(ref Vertex v, Proj4Projection proj)
        {
            try
            {
                proj.Open();

                v = proj.ConvertToExternalProjection(v.X, v.Y);
            }
            finally
            {
                proj.Close();
            }
        }

        /// <summary>
        /// Projects a RectangleShape to the specified Proj4Projection's ExternalProjection.
        /// </summary>
        /// <param name="rec">RectangleShape to project.</param>
        /// <param name="proj">Proj4Projection to use.</param>
        /// <returns>A RectangleShape projected to the specified Proj4Projection's ExternalProjection.</returns>
        public static RectangleShape ProjectRectangleShape(RectangleShape rec, Proj4Projection proj)
        {
            RectangleShape r;
            try
            {
                proj.Open();

                r = proj.ConvertToExternalProjection(rec);
            }
            finally
            {
                proj.Close();
            }

            return r;
        }

        /// <summary>
        /// Projects a RectangleShape to the specified Proj4Projection's ExternalProjection.
        /// </summary>
        /// <param name="rec">RectangleShape to project.</param>
        /// <param name="proj">Proj4Projection to use.</param>
        public static void ProjectRectangleShape(ref RectangleShape rec, Proj4Projection proj)
        {
            try
            {
                proj.Open();

                rec = proj.ConvertToExternalProjection(rec);
            }
            finally
            {
                proj.Close();
            }
        }

        #endregion

        /// <summary>
        /// Returns an array of doubles, representing the points making up a RectangleShape, in the order Left, Bottom, Right, Top. Used for creating an OpenLayers Bounds object client side.
        /// </summary>
        /// <param name="rec">The RectangleShape to get the array from.</param>
        /// <returns>Array of doubles, representing the points making up a rectangle shape, in the order Left, Bottom, Right, Top.</returns>
        public static double[] GetArrayFromRectangleShape(RectangleShape rec)
        {
            return new double[] { rec.LowerLeftPoint.X, rec.LowerLeftPoint.Y, rec.UpperRightPoint.X, rec.UpperRightPoint.Y };
        }

        /// <summary>
        ///  Returns an array of doubles, representing the points making up a RectangleShape, in the order Left, Bottom, Right, Top, projected to the specified Proj4Projection's ExternalProjection. Used for creating an OpenLayers Bounds object client side.
        /// </summary>
        /// <param name="rec">The RectangleShape to get the array from.</param>
        /// <param name="proj">Proj4Projection to use.</param>
        /// <returns>>Array of doubles, representing the points making up a rectangle shape, in the order Left, Bottom, Right, Top, projected to the specified Proj4Projection's ExternalProjection.</returns>
        public static double[] GetArrayForProjectedRectangleShape(RectangleShape rec, Proj4Projection proj)
        {
            return GetArrayFromRectangleShape(ProjectRectangleShape(rec, proj));
        }

        /// <summary>
        /// Gets the RectangleShape representing the DefaultExtent property projected to the specified projection.
        /// </summary>
        /// <param name="proj">The projection to use.</param>
        /// <returns>The projected RectangleShape for the DefaultExtent.</returns>
        public static RectangleShape GetProjectedDefaultExtent(Proj4Projection proj)
        {
            return ProjectRectangleShape(DefaultExtent, proj);
        }

        /// <summary>
        /// Returns an array of doubles, representing the points making up a RectangleShape, in the order Left, Bottom, Right, Top, for the DefaultExtent property. Used for creating an OpenLayers Bounds object client side.
        /// </summary>
        /// <param name="proj"></param>
        /// <returns>Array of doubles, representing the points making up the DefaultExtent, in the order Left, Bottom, Right, Top, projected to the specified Proj4Projection's ExternalProjection.</returns>
        public static double[] GetArrayForProjectedDefaultExtent(Proj4Projection proj)
        {
            return GetArrayFromRectangleShape(GetProjectedDefaultExtent(proj));
        }
    }
}
